/**
 * https://invent.kde.org/plasma/plasma-browser-integration/-/blob/master/dbus/org.mpris.MediaPlayer2.Player.xml?ref_type=heads
 */

import dbus from "dbus-next";
import RPC, { Presence } from "discord-rpc";

const CLIENT_ID = "1237419378383781949";

let rpcClient = new RPC.Client({ transport: "ipc" });
rpcClient.login({ clientId: CLIENT_ID, scopes: [] })

type PlaybackData = {
    isDeezer?: boolean,
    artUrl?: string,
    album?: string,
    artist?: string[],
    title?: string,
    length?: number,
    position?: number,
    playing: boolean,
    paused?: boolean,
}

export enum ActivityType {
    Playing = 0,
    Listening = 2,
    Watching = 3,
    Competing = 5,
}

let playbackData: PlaybackData = { playing: false };

const bus: dbus.MessageBus = dbus.sessionBus();
let obj: dbus.ProxyObject;
let player: dbus.ClientInterface;
let props: dbus.ClientInterface;

async function run() {
    while (true) {
        try {
            obj = await bus.getProxyObject("org.kde.plasma.browser_integration", "/org/mpris/MediaPlayer2");
            player = obj.getInterface("org.mpris.MediaPlayer2.Player");
            props = obj.getInterface("org.freedesktop.DBus.Properties");
            break;
        } catch(e) {
            console.log("Failed to connect: " + e + "\nRetrying in 5000ms");
            await new Promise(r => setTimeout(r, 5000));
        }
    }

    async function fetchData() {
        parseData("Metadata", (await props.Get("org.mpris.MediaPlayer2.Player", "Metadata")).value);
        parseData("PlaybackStatus", (await props.Get("org.mpris.MediaPlayer2.Player", "PlaybackStatus")).value);
        parseData("Position", (await props.Get("org.mpris.MediaPlayer2.Player", "Position")).value);
    }

    function parseData(prop: string, data: any) {
        switch(prop) {
            case 'Metadata': {
                if (!data?.["xesam:url"]?.value?.startsWith("https://www.deezer.com/")) {
                    playbackData = { playing: false, isDeezer: false };
                    return;
                }
            
                playbackData.isDeezer = true;
                playbackData.artUrl = data?.["mpris:artUrl"]?.value;
                playbackData.length = Number(data?.["mpris:length"]?.value) / 1000;
                playbackData.album = data?.["xesam:album"]?.value;
                playbackData.artist = data?.["xesam:artist"]?.value;
                playbackData.title = data?.["xesam:title"]?.value;
                break;
            }
            case 'PlaybackStatus': {
                playbackData.playing = data == "Playing";
                playbackData.paused = data == "Paused";
                break;
            }
            case 'Position': {
                playbackData.position = Number(data) / 1000;
                break;
            }
            default: {
                console.debug("Received data update for", prop, "- ignoring")
                return;
            }
        }
        console.log("Processed update for prop", prop);
    }

    await fetchData();

    let interval: any;
    await setPresence(true);

    async function setPresence(manual: boolean = false) {
        if (manual) {
            if (typeof interval != "undefined") clearInterval(interval);
            interval = setInterval(setPresence, 15000); // Regularly update to make sure Discord doesn't forget about us for some reason
        }

        console.log("Setting presence");
        if ((!playbackData.playing && !playbackData.paused) || !playbackData.isDeezer) {
            await rpcClient.clearActivity();
        }
        else {
            const presence: Presence = {
                details: playbackData.title?.slice(0, 128) ?? "(Unknown song)",
                state: playbackData.artist?.join(", ").slice(0, 128) ?? "(Unknown artist)",
                endTimestamp: playbackData.playing ? Date.now() + (playbackData.length! - playbackData.position!) : undefined,
                largeImageKey: playbackData.artUrl ?? "deezer_logo",
                largeImageText: `${playbackData.artist} - ${playbackData.title}`.slice(0, 128),
                smallImageKey: playbackData.playing ? "icon_play" : playbackData.paused ? "icon_pause" : undefined,
                smallImageText: playbackData.playing ? "Playing" : playbackData.paused ? "Paused" : undefined,
                buttons: [
                    { label: 'Play on Deezer', url: `https://www.deezer.com/search/${encodeURIComponent(`${playbackData.artist} - ${playbackData.title}`)}` },
                ],
                type: ActivityType.Listening
            } as any;
            await rpcClient.setActivity(presence);
        }
    }

    setInterval(() => playbackData.position && playbackData.position++, 1);

    props.on('PropertiesChanged', (iface: string, changed: { [key: string]: any }[], invalidated: any[]) => {
        for (let [prop, data] of Object.entries(changed)) {
            parseData(prop, data.value);
        }
        setPresence(true);
    });

    // I have no idea what event to listen to to detect when the browser
    // exits, so we poll it and enter the reconnect loop on error
    async function pollStatus() {
        try {
            await props.Get("org.mpris.MediaPlayer2.Player", "PlaybackStatus");
            setTimeout(pollStatus, 500);
        } catch(e) {
            console.log("Failed to poll status: " + e + "\nConsidering this connection dead");
            if (interval) clearInterval(interval);
            props.removeAllListeners();
            playbackData = { playing: false };
            await rpcClient.clearActivity();

            // Restart everything on a new stack
            setTimeout(run, 0);
        }
    }
    await pollStatus();
}

run();
